require 'singleton'
require_relative 'service'
require_relative 'session'

module Zzlink
  # A zzLink user.
  class User
    attr_accessor :id, :name, :avatar_url, :home_url, :email
    attr_accessor :access_token, :token_type, :oauth_scope, :oauth_id, :oauth_provider
    attr_accessor :created_at, :updated_at

    # New a User object.
    # @param attributes [Hash] attributes of the new user.
    def initialize(attributes = {})
      from_hash(attributes)
    end

    # Assign attributes from a Hash.
    # @param attributes [Hash] attributes.
    def from_hash(attributes)
      attributes.each_pair do |k, v|
        setter = case k.to_sym
                 when :avatarURL then :avatar_url=
                 when :homeURL then :home_url=
                 when :accessToken then :access_token=
                 when :tokenType then :token_type=
                 when :oauthScope then :oauth_scope=
                 when :oauthId then :oauth_id=
                 when :oauthProvider then :oauth_provider=
                 when :createdAt then :created_at=
                 when :updatedAt then :created_at=
                 else "#{k}=".to_sym
                 end
        send(setter, v) if respond_to?(setter)
      end
    end

    def to_hash
      hash = {}
      hash['id'] = id unless id.nil?
      hash['name'] = name unless name.nil?
      hash['avatarURL'] = avatar_url unless avatar_url.nil?
      hash['homeURL'] = home_url unless home_url.nil?
      hash['email'] = email unless email.nil?
      hash['accessToken'] = access_token unless access_token.nil?
      hash['tokenType'] = token_type unless token_type.nil?
      hash['oauthScope'] = oauth_scope unless oauth_scope.nil?
      hash['oauthId'] = oauth_id unless oauth_id.nil?
      hash['oauthProvider'] = oauth_provider unless oauth_provider.nil?
      hash['createdAt'] = created_at unless created_at.nil?
      hash['updatedAt'] = updated_at unless updated_at.nil?
      hash
    end
  end

  # The zzUser service.
  class Users < Service
    include Singleton

    def name
      'zzUsers'
    end

    def version
      'v1.0'
    end

    # Create a user. If :oauth_provider and :oauth_id already exist, the existing user will be updated.
    # @param attributes [Hash] attributes of the new user.
    # @return [User] the created user.
    # @raise [Error] in case of failure.
    def create_user(attributes)
      user = User.new(attributes)
      resp = post('/users', user.to_hash)
      user.from_hash(JSON.parse(resp))
      user
    end

    # Create or reset a session.
    # @param attributes [Hash] attributes of the new session. If the combination of user_id and client_id confilicts, the existing session will be reset.
    # @return [Session] the created or updated session.
    # @raise [Error] in case of failure.
    def create_session(attributes)
      session = Session.new(attributes)
      resp = post('/sessions', session.to_hash)
      session.from_hash(JSON.parse(resp))
      session
    end

    # Find a session by its token.
    # @param token [String] the token.
    # @return [Session, nil] the found session or nil.
    def find_session_by_token(token)
      begin
        resp = get("/sessions/#{token}")
        Session.new(JSON.parse(resp))
      rescue
        nil
      end
    end

    # Delete a session by its token.
    # @param token [String] the token.
    # @return [Boolean] the found session or nil.
    def delete_session_by_token(token)
      begin
        delete("/sessions/#{token}")
        true
      rescue
        false
      end
    end
  end
end
